@@ -2,13 +2,10 @@
 // SPDX-License-Identifier: Apache-2.0
 // SPDX-License-Identifier: MIT
 
-#[cfg(target_os = "linux")]
+#[cfg(not(target_os = "macos"))]
 fn main() {}
 
 #[cfg(target_os = "macos")]
 fn main() {
   println!("cargo:rustc-link-lib=framework=WebKit");
 }
-
-#[cfg(target_os = "windows")]
-fn main() {}
