@@ -16,8 +16,8 @@ fn main() -> wry::Result<()> {
   let window = WindowBuilder::new()
     .with_title("Hello World")
     .build(&event_loop)?;
-  let _webview = WebViewBuilder::new(window)?
-    .with_url("https://tauri.studio")?
+  let webview = WebViewBuilder::new(window)?
+    .with_url("https://html5test.com")?
     .build()?;
 
   event_loop.run(move |event, _, control_flow| {
@@ -29,7 +29,9 @@ fn main() -> wry::Result<()> {
         event: WindowEvent::CloseRequested,
         ..
       } => *control_flow = ControlFlow::Exit,
-      _ => (),
+      _ => {
+        dbg!(webview.window().inner_size());
+      }
     }
   });
 }
